#include <stdio.h>
#include <regex.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <unistd.h>
#include <list>
#include <vector>


static void show_usage(FILE *fp) {
	fprintf(fp,"Name\n");
	fprintf(fp,"    grepcut - combined grep and cut tool\n");
	fprintf(fp,"Synopsis\n");
	fprintf(fp,"    grepcut [OPTIONS] [FILE...]\n");
	fprintf(fp,"Options\n");
	fprintf(fp,"    -E\n");	
	fprintf(fp,"        Interpret PATTERN as an extended regular expression.\n");
	fprintf(fp,"    -F\n");
	fprintf(fp,"        Interpret PATTERN as a list  of fixed strings.\n");
	fprintf(fp,"    -G\n");
	fprintf(fp,"        Interpret PATTERN as a basic regular expression.\n");
	fprintf(fp,"    -e PATTERN\n");
	fprintf(fp,"        Use PATTERN as the pattern.\n");
	fprintf(fp,"    -f FILE\n");
	fprintf(fp,"        Obtain patterns from FILE, one per line.\n");
	fprintf(fp,"    -v\n");
	fprintf(fp,"        Invert the sense of matching, to select non-matching lines.\n");
	fprintf(fp,"    -i\n");
	fprintf(fp,"        Ignore case distinctions in both the PATTERN and the input files.\n");
	fprintf(fp,"    -h\n");
	fprintf(fp,"        Suppress the prefixing of file names on output.\n");
	fprintf(fp,"    -H\n");
	fprintf(fp,"        Print the file name for each match.\n");
	fprintf(fp,"    -c LIST\n");
	fprintf(fp,"        Select only these characters.\n");
	fprintf(fp,"    -d DELIM\n");
	fprintf(fp,"        Use DELIM instead of TAB for field delimiter.\n");
	fprintf(fp,"    -l LIST\n");
	fprintf(fp,"        Select only these fields.\n");
	fprintf(fp,"    -w LIST\n");
	fprintf(fp,"        Select only these fields (multiple delimiters are treated as one).\n");
}

struct Range {
	size_t start;
	size_t end;
};

struct CutOperation {
	enum mode_t {
		mode_chars,
		mode_fields,
		mode_words
	};
	mode_t mode;
	char delimiter;
	std::list<Range> ranges;
};

static std::list<CutOperation> cuts;

enum grep_mode_t {
	grep_mode_fixed,
	grep_mode_basic,
	grep_mode_extended
};
static grep_mode_t grep_mode = grep_mode_basic;
static bool complement_match=false;
static bool ignore_case=false;
static bool with_filename=true;

struct Pattern {
	std::string original_s;
	std::string s;
	regex_t reg;
};
static std::list<Pattern> patterns;

const char *current_filename=0;


static void read_patterns(const char *filename) {
	FILE *fp = fopen(filename,"r");
	if(!fp) {
		perror(filename);
		exit(2);
	}
	char line[1024];
	while(fgets(line,sizeof(line),fp)) {
		char *s = strchr(line,'\n');
		if(s)
			*s = '\0';
		Pattern p;
		p.original_s = s;
		patterns.push_back(p);
	}
	fclose(fp);
}


static void parse_range(std::list<Range> *ranges, char *list) {
	char *ss1;
	for(char *s = strtok_r(list,",",&ss1); s; s=strtok_r(0,",",&ss1)) {
		char *endptr;
		unsigned long range_start = strtoul(s,&endptr,10);
		if(endptr && *endptr!='\0' && *endptr!='-') {
			fprintf(stderr,"Invalid range: %s\n",s);
			exit(98);
		}
		if(range_start<1) {
			fprintf(stderr,"Invalid range start (ranges are 1-based): %s\n",s);
			exit(98);
		}
		unsigned long range_end;
		if(endptr && *endptr=='-') {
			endptr++;
			if(*endptr) {
				range_end = strtoul(endptr,&endptr,10);
				if(endptr && *endptr) {
					fprintf(stderr,"Invalid range: %s\n",s);
					exit(98);
				}
				if(range_start>range_end) {
					fprintf(stderr,"Invalid range: %s\n",s);
					exit(98);
				}
			} else
				range_end=4095; //"infinity"
		} else
			range_end = range_start;
		Range r;
		r.start = range_start;
		r.end = range_end;
		ranges->push_back(r);
	}
}


static void compile_patterns() {
	if(grep_mode==grep_mode_fixed) {
		//escape special characters
		for(std::list<Pattern>::iterator iter=patterns.begin(); iter!=patterns.end(); ++iter) {
			std::string new_s;
			for(size_t i=0; i<iter->original_s.size(); i++) {
				char c = iter->original_s[i];
				if(c=='\\' || c=='.' || c=='*' || c=='[' || c=='^' || c=='$')
					new_s += '\\';
				new_s += c;
			}
			iter->s = new_s;
		}
	} else {
		for(std::list<Pattern>::iterator iter=patterns.begin(); iter!=patterns.end(); ++iter)
			iter->s = iter->original_s;
	}
	
	int cflags = 0;
	if(grep_mode==grep_mode_extended)
		cflags |= REG_EXTENDED;
	if(ignore_case)
		cflags |= REG_ICASE;
	
	for(std::list<Pattern>::iterator iter=patterns.begin(); iter!=patterns.end(); ++iter) {
		int rc = regcomp(&iter->reg,iter->s.c_str(),cflags);
		if(rc!=0) {
			char errmsg[256];
			(void)regerror(rc,&iter->reg,errmsg,sizeof(errmsg));
			fprintf(stderr,"regcomp() error on %s: %s\n",iter->s.c_str(),errmsg);
			exit(98);
		}
	}
}

static void decompile_patterns() {
	for(std::list<Pattern>::iterator iter=patterns.begin(); iter!=patterns.end(); ++iter)
		regfree(&iter->reg);
}


static bool line_matches(const char *line) {
	if(patterns.empty())
		return true;
	bool match_found = false;
	for(std::list<Pattern>::iterator iter=patterns.begin(); iter!=patterns.end(); ++iter) {
		int rc = regexec(&iter->reg, line, 0,0, 0);
		if(rc==0) {
			match_found=true;
			break;
		}
	}
	
	if(!complement_match)
		return match_found;
	else
		return !match_found;
}


static std::string do_cut_chars(const CutOperation &co, const std::string &str) {
	std::string result;
	for(std::list<Range>::const_iterator iter=co.ranges.begin(); iter!=co.ranges.end(); ++iter) {
		const Range &r = *iter;
		if(r.start-1>=str.size())
			continue;
		result += str.substr(r.start-1,r.end-r.start+1);
	}
	return result;
}

struct Section {
	const char *start;
	size_t len;
	Section() {}
	Section(const char *start_, size_t len_) : start(start_), len(len_) {}
};

static std::string do_cut_fields(const CutOperation &co, const std::string &str) {
	//todo: more efficient
	std::vector<Section> fields;
	
	//split into fields
	size_t field_start=0;
	for(size_t i=0; i<str.size(); i++) {
		if(str[i]==co.delimiter) {
			fields.push_back(Section(str.data()+field_start,i-field_start));
			field_start=i+1;
		}
	}
	if(field_start<str.size())
		fields.push_back(Section(str.data()+field_start,str.size()-field_start));
	
	//output
	std::string result;
	for(std::list<Range>::const_iterator iter=co.ranges.begin(); iter!=co.ranges.end(); ++iter) {
		const Range &r = *iter;
		std::string range_result;
		for(size_t i=r.start; i<=r.end && i<fields.size(); i++) {
			if(i-1<fields.size()) {
				if(!range_result.empty())
					range_result += co.delimiter;
				range_result += std::string(fields[i-1].start,fields[i-1].len);
			}
		}
		if(!range_result.empty() && !result.empty()) {
			result += co.delimiter;
		}
		result += range_result;
	}
	return result;
}


static std::string do_cut_words(const CutOperation &co, const std::string &str) {
	//todo: more efficient
	std::vector<Section> fields;
	
	//split into fields
	for(size_t i=0; i<str.size(); ) {
		while(i<str.size() && str[i]==co.delimiter)
			i++;
		if(i==str.size())
			break;
		size_t field_start = i;
		while(i<str.size() && str[i]!=co.delimiter)
			i++;
		fields.push_back(Section(str.data()+field_start,i-field_start));
	}
	
	//output
	std::string result;
	for(std::list<Range>::const_iterator iter=co.ranges.begin(); iter!=co.ranges.end(); ++iter) {
		const Range &r = *iter;
		std::string range_result;
		for(size_t i=r.start; i<=r.end && i<fields.size(); i++) {
			if(i-1<fields.size()) {
				if(!range_result.empty())
					range_result += co.delimiter;
				range_result += std::string(fields[i-1].start,fields[i-1].len);
			}
		}
		if(!range_result.empty() && !result.empty()) {
			result += co.delimiter;
		}
		result += range_result;
	}
	return result;
}


static std::string do_cut(const CutOperation &co, const std::string &str) {
	switch(co.mode) {
		case CutOperation::mode_chars:
			return do_cut_chars(co,str);
		case CutOperation::mode_fields:
			return do_cut_fields(co,str);
		case CutOperation::mode_words:
			return do_cut_words(co,str);
		default:
			fprintf(stderr,"Internal error\n");
			exit(10);
	}
}


static void cut_and_output_line(const char *line) {
	if(cuts.empty()) {
		fputs(line,stdout);
		putc('\n',stdout);
		return;
	}
	std::string tmp(line);
	for(std::list<CutOperation>::const_iterator iter=cuts.begin(); iter!=cuts.end(); ++iter) {
		tmp = do_cut(*iter,tmp);
	}
	if(with_filename && current_filename) {
		fputs(current_filename,stdout);
		fputc(':',stdout);
	}
	fputs(tmp.c_str(),stdout);
	putc('\n',stdout);
}


static void process_line(const char *line) {
	if(line_matches(line))
		cut_and_output_line(line);
}


static void process_stdin() {
	char line[4096];
	while(fgets(line,sizeof(line),stdin)) {
		char *nl=strchr(line,'\n');
		if(nl) *nl='\0';
		process_line(line);
	}
}


static void process_file(const char *filename) {
	FILE *fp = fopen(filename,"r");
	if(!fp) {
		perror(filename);
		exit(2);
	}
	current_filename = filename;
	char line[4096];
	while(fgets(line,sizeof(line),fp)) {
		char *nl=strchr(line,'\n');
		if(nl) *nl='\0';
		process_line(line);
	}
	fclose(fp);
}


static void process_input(int argc, char **argv, int start_arg_index) {
	if(start_arg_index==argc)
		process_stdin();
	else {
		for(int i=start_arg_index; i<argc; i++)
			process_file(argv[i]);
	}
}


int main(int argc, char **argv) {
	int c;
	char next_delimiter='\t';
	while((c=getopt(argc,argv,"EFGe:f:vihHc:d:l:w:"))!=EOF) {
		switch(c) {
			case 'E':
				grep_mode = grep_mode_extended;
				break;
			case 'F':
				grep_mode = grep_mode_fixed;
				break;
			case 'G':
				grep_mode = grep_mode_basic;
				break;
			case 'e': {
				Pattern p;
				p.original_s = optarg;
				patterns.push_back(p);
				break;
			}
			case 'f':
				read_patterns(optarg);
				break;
			case 'v':
				complement_match = true;
				break;
			case 'i':
				ignore_case = true;
				break;
			case 'h':
				with_filename = false;
				break;
			case 'H':
				with_filename = true;
				break;
			case 'c': {
				CutOperation co;
				co.mode = CutOperation::mode_chars;
				co.delimiter = next_delimiter;
				parse_range(&co.ranges,optarg);
				cuts.push_back(co);
				break;
			}
			case 'd':
				if(strlen(optarg)!=1) {
					fprintf(stderr,"Delimiter must be exactly 1 character long\n");
					return 98;
				}
				next_delimiter = *optarg;
				break;
			case 'l': {
				CutOperation co;
				co.mode = CutOperation::mode_fields;
				co.delimiter = next_delimiter;
				parse_range(&co.ranges,optarg);
				cuts.push_back(co);
				break;
			}
			case 'w': {
				CutOperation co;
				co.mode = CutOperation::mode_words;
				co.delimiter = next_delimiter;
				parse_range(&co.ranges,optarg);
				cuts.push_back(co);
				break;
			}
			default:
				show_usage(stderr);
				return 99;
		}
	}
	
	compile_patterns();
	
	process_input(argc,argv,optind);
	
	decompile_patterns();
	
	return 0;
}
